import * as THREE from 'https://cdnjs.cloudflare.com/ajax/libs/three.js/r126/three.module.js';
import { OrbitControls } from './OrbitControls.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        this.domWebGL = document.createElement('div');
        document.body.appendChild(this.domWebGL);

        let scene = new THREE.Scene();
        let renderer = new THREE.WebGLRenderer();

        renderer.setClearColor(0x000000, 1.0);

        this.domWebGL.appendChild(renderer.domElement);
        window.onresize = this.resize.bind(this);

        this.renderer = renderer;
        this.scene = scene;

        this._setupModel();
        this._setupLights()
        this._setupCamera();
        this._setupControl();

        this.resize();
    }

    _setupControl() { //!
        const control = new OrbitControls(this.camera, this.renderer.domElement);
        control.enableDamping = true;
        control.enablePan = false;
        control.minDistance = 1.2;
        control.maxDistance = 4;
        control.update();

        this._control = control;
    }


    _getPosFromLatLonRad(lat, lon, radius) {
        var phi = (90 - lat) * (Math.PI / 180)
        var theta = (lon + 180) * (Math.PI / 180)
    
        let x = -((radius) * Math.sin(phi)*Math.cos(theta))
        let z = ((radius) * Math.sin(phi)*Math.sin(theta))
        let y = ((radius) * Math.cos(phi))
    
        return {x,y,z}
    }

    _createCurveMesh(p1, p2, material) {
        let v1 = new THREE.Vector3(p1.x, p1.y, p1.z)
        let v2 = new THREE.Vector3(p2.x, p2.y, p2.z)
    
        let points = []
    
        const step = 50;
        for(let i=0; i<step+1; i++) {
            let p = new THREE.Vector3().lerpVectors(v1, v2, i/step)
            p.normalize();
            p.multiplyScalar(1 + 0.2*Math.sin(Math.PI*i/step))
            points.push(p)
        }
    
        const path = new THREE.CatmullRomCurve3(points)

        const geometry = new THREE.TubeGeometry(path, step, 0.002, 5, false)
        //const material = new THREE.MeshBasicMaterial({ color: 0x0000ff })
        const mesh = new THREE.Mesh(geometry, material)
        return mesh;
    }

    _createCircleMesh(pos, color) {
        const geometry = new THREE.TorusGeometry(0.01, 0.001, 16, 32 );
        const material = new THREE.MeshBasicMaterial( { color } );
        const torus = new THREE.Mesh( geometry, material );

        torus.position.set(pos.x, pos.y, pos.z);
        torus.lookAt(-pos.x, -pos.y, -pos.z)

        return torus;
    }

_setupModel() {
    // EARTH
    const geometry = new THREE.SphereBufferGeometry(1, 30, 30);
    const material = new THREE.MeshBasicMaterial({ 
        map: new THREE.TextureLoader().load("./earth.jpg") 
    });
    const earth = new THREE.Mesh(geometry, material);
    this.scene.add(earth);
    // .

    // SEOUL
    const seoul_lat = 37.1;
    const seoul_lng = 126.58;
    const seoul_pos = this._getPosFromLatLonRad(seoul_lat, seoul_lng, 1);
    const seoul = this._createCircleMesh(seoul_pos, 0xff0000);
    this.scene.add(seoul);
    // .

    // NEWYORK //
    const newyork_lat = 40.60857;
    const newyork_lng = -74.01559;
    const newyork_pos = this._getPosFromLatLonRad(newyork_lat, newyork_lng, 1);

    const newyork = this._createCircleMesh(newyork_pos, 0xffff00);
    this.scene.add(newyork);
    // .

    this.scalingObjects = [seoul, newyork];

    this._curve_material = new THREE.ShaderMaterial({
        extensions: {
            //derivatives: "#extension GL_OES_standard_derivatives : enable"
        },
        uniforms: {
            time: { value: 0 },
        },
        side: THREE.DoubleSide,
        transparent: true,
        //blending: THREE.AdditiveBlending,
        //depthTest: false,

        vertexShader: `
            varying vec2 vUv;
            void main() {
                vUv = uv;
                gl_PointSize = 5.0;
                vec4 modelViewPosition = modelViewMatrix * vec4(position, 1.0);
                gl_Position = projectionMatrix * modelViewPosition;
            }
        `,
        
        fragmentShader: `
            varying vec2 vUv;
            uniform float time;
            void main() {
                float a = vUv.x;
                float t = time / 2.0;
                float d = 0.2;
                if(t < a) discard;
                if((t-d) > a) discard;
                gl_FragColor = vec4(1, 1, 1, 1);
            }
        `
    });

    this._curveMesh = this._createCurveMesh(seoul_pos, newyork_pos, this._curve_material);
    this.scene.add(this._curveMesh);
}

    update(time) {
        if(this._previousTime === undefined) {
            this._previousTime = 0;
        }
        
        if(time - this._previousTime > 2500) {
            this._previousTime = time;
        }

        const t = (time - this._previousTime) / 1000.0;
        this._curve_material.uniforms.time.value = t;

        this.scalingObjects.forEach((obj, ndx) => {
            obj.scale.set(t,t,t);
        });
        
    }

    _setupLights() {
        const color = 0xFFFFFF;
        const intensity = 1;

        const light1 = new THREE.PointLight(color, intensity);
        light1.position.set(10, 0, 10);
        this.scene.add(light1);

        const light2 = new THREE.PointLight(color, intensity);
        light2.position.set(10, 0, -10);
        this.scene.add(light2);   
    }

    _setupCamera() {
        let camera = new THREE.PerspectiveCamera(75, window.innerWidth / window.innerHeight, 0.1, 1000);

        camera.position.set(0, 0, 3);
        //camera.up.set(0, 0, 1);
        camera.lookAt(0, 0, 0);

        this.scene.add(camera);
        this.camera = camera;
    }

    render(time) {
        this.update(time);
        this._control.update();
        this.renderer.render(this.scene, this.camera);
        requestAnimationFrame(this.render.bind(this));
    }

    resize() {
        let camera = this.camera;
        let renderer = this.renderer;

        renderer.setPixelRatio(window.devicePixelRatio);
        renderer.setSize(window.innerWidth, window.innerHeight);

        camera.aspect = window.innerWidth / window.innerHeight;
        camera.updateProjectionMatrix();
    }
}

window.onload = function () {
    (new App()).render();
}